"""
Unit testing for parsing modules with tasks created via the PrismTask class.

Table of Contents:
- Imports
- Test case directory and paths
- Test case class definition
"""

###########
# Imports #
###########

# Standard library imports
import os
from pathlib import Path
import unittest

# Prism imports
import prism.exceptions
import prism.parsers.ast_parser as ast_parser


#################################
# Test case directory and paths #
#################################

# Directory containing all prism_project.py test cases
TEST_CASE_WKDIR = os.path.dirname(__file__)
MODULE_TEST_CASES = Path(TEST_CASE_WKDIR) / 'test_modules'

# Test cases related to imports
ONE_PRISM_TASK = Path('one_prism_task.py')
NO_PRISM_TASK = Path('no_prism_task.py')
MULTIPLE_PRISM_TASKS = Path('multiple_prism_tasks.py')
DIFF_IMPORT_STRUCTURE = Path('diff_import_structure.py')
OTHER_CLASSES = Path('other_classes.py')
IF_NAME_MAIN = Path('if_name_main.py')
TASKS_REFS = Path('tasks_refs.py')
TASK_WITH_TARGET = Path('task_with_target.py')
BAD_RUN_EXTRA_ARG = Path('bad_run_extra_arg.py')
BAD_RUN_MISSING_ARG = Path('bad_run_missing_arg.py')
BAD_RUN_NO_TASKS = Path('bad_run_no_tasks.py')
NO_RUN_FUNC = Path('no_run_func.py')


##############################
# Test case class definition #
##############################

class TestModuleClassParsing(unittest.TestCase):

    def test_one_prism_task(self):
        """
        Test behavior of parser with just one Prism task
        """
        # Prism task
        parser = ast_parser.AstParser(ONE_PRISM_TASK, MODULE_TEST_CASES)

        # Prism task name
        prism_task_class = parser.get_prism_task_node(parser.classes, parser.bases)
        prism_task_name = prism_task_class.name

        # Run function
        run_func = parser.get_run_func(prism_task_class)
        run_func_args = parser.get_func_args(run_func)

        # Number of prism tasks
        num_prism_tasks = parser.get_num_prism_task_classes(parser.bases)

        self.assertEqual("NormalPrismTask", prism_task_name)
        self.assertEqual(1, num_prism_tasks)
        self.assertEqual(['self', 'tasks', 'hooks'], run_func_args)

        # Calling `parse` shouldn't throw an error
        self.assertEqual([], parser.parse())

    def test_no_prism_task(self):
        """
        Test behavior of parser when module has no Prism task
        """
        parser = ast_parser.AstParser(NO_PRISM_TASK, MODULE_TEST_CASES)

        # Prism task name
        prism_task_class = parser.get_prism_task_node(parser.classes, parser.bases)

        # Number of prism tasks
        num_prism_tasks = parser.get_num_prism_task_classes(parser.bases)

        self.assertIsNone(prism_task_class)
        self.assertEqual(0, num_prism_tasks)

        # Calling `parse` should throw an error
        with self.assertRaises(prism.exceptions.ParserException) as cm:
            parser.parse()
        expected_msg = f"no PrismTask in `{str(NO_PRISM_TASK)}`"
        self.assertEqual(expected_msg, str(cm.exception))

    def test_multiple_prism_task(self):
        """
        Test behavior of parser when module has multiple Prism tasks
        """
        parser = ast_parser.AstParser(MULTIPLE_PRISM_TASKS, MODULE_TEST_CASES)

        # Number of prism tasks
        num_prism_tasks = parser.get_num_prism_task_classes(parser.bases)
        self.assertEqual(2, num_prism_tasks)

        # Prism task name -- it should return the first one
        with self.assertRaises(prism.exceptions.RuntimeException) as cm:
            _ = parser.get_prism_task_node(
                parser.classes, parser.bases
            )
        expected_msg = f"too many PrismTasks in `{str(MULTIPLE_PRISM_TASKS)}`"
        self.assertEqual(expected_msg, str(cm.exception))

    def test_diff_import_structure(self):
        """
        Different PrismTask import structure (i.e., import prism.task.PrismTask) should
        not affect the behavior of the parser
        """
        # Prism task
        parser = ast_parser.AstParser(DIFF_IMPORT_STRUCTURE, MODULE_TEST_CASES)

        # Prism task name
        prism_task_class = parser.get_prism_task_node(parser.classes, parser.bases)
        prism_task_name = prism_task_class.name

        # Run function
        run_func = parser.get_run_func(prism_task_class)
        run_func_args = parser.get_func_args(run_func)

        # Number of prism tasks
        num_prism_tasks = parser.get_num_prism_task_classes(parser.bases)

        self.assertEqual("DiffImportStructure", prism_task_name)
        self.assertEqual(1, num_prism_tasks)
        self.assertEqual(['self', 'tasks', 'hooks'], run_func_args)

        # Calling `parse` shouldn't throw an error
        self.assertEqual([], parser.parse())

    def test_other_classes(self):
        """
        Presence of other, non-PrismTask classes should not affect the behavior of the
        parser
        """
        # Prism task
        parser = ast_parser.AstParser(OTHER_CLASSES, MODULE_TEST_CASES)

        # Prism task name
        prism_task_class = parser.get_prism_task_node(parser.classes, parser.bases)
        prism_task_name = prism_task_class.name

        # Run function
        run_func = parser.get_run_func(prism_task_class)
        run_func_args = parser.get_func_args(run_func)

        # Number of prism tasks
        num_prism_tasks = parser.get_num_prism_task_classes(parser.bases)

        self.assertEqual("OnlyPrismTask", prism_task_name)
        self.assertEqual(1, num_prism_tasks)
        self.assertEqual(['self', 'tasks', 'hooks'], run_func_args)

        # Calling `parse` shouldn't throw an error
        self.assertEqual([], parser.parse())

    def test_task_with_target(self):
        """
        Presence of target does not affect parser behavior
        """
        # Prism task
        parser = ast_parser.AstParser(TASK_WITH_TARGET, MODULE_TEST_CASES)

        # Prism task name
        prism_task_class = parser.get_prism_task_node(parser.classes, parser.bases)
        prism_task_name = prism_task_class.name

        # Run function
        run_func = parser.get_run_func(prism_task_class)
        run_func_args = parser.get_func_args(run_func)

        # Number of prism tasks
        num_prism_tasks = parser.get_num_prism_task_classes(parser.bases)

        self.assertEqual("TaskWithTarget", prism_task_name)
        self.assertEqual(1, num_prism_tasks)
        self.assertEqual(['self', 'tasks', 'hooks'], run_func_args)

        # Calling `parse` shouldn't throw an error
        self.assertEqual([Path('hello.py'), Path('world.py')], parser.parse())

        # Get target
        targets = parser.get_targets(prism_task_class, run_func)
        expected_targets = "os.path.join(os.getcwd(), 'temp')"
        self.assertEqual(targets, expected_targets)

    def test_tasks_refs(self):
        """
        Test behavior of parse when there are mod references
        """
        # Prism task
        parser = ast_parser.AstParser(TASKS_REFS, MODULE_TEST_CASES)

        # Prism task name
        prism_task_class = parser.get_prism_task_node(parser.classes, parser.bases)
        prism_task_name = prism_task_class.name

        # Run function
        run_func = parser.get_run_func(prism_task_class)
        run_func_args = parser.get_func_args(run_func)

        # Number of prism tasks
        num_prism_tasks = parser.get_num_prism_task_classes(parser.bases)

        self.assertEqual("TasksRefs", prism_task_name)
        self.assertEqual(1, num_prism_tasks)
        self.assertEqual(['self', 'tasks', 'hooks'], run_func_args)

        # Calling `parse` shouldn't throw an error
        expected_tasks = [
            Path('func_0.py'),
            Path('func_1.py'),
            Path('hello.py'),
            Path('world.py')
        ]
        self.assertEqual(sorted(expected_tasks), sorted(parser.parse()))

    def test_if_name_main(self):
        """
        If a module contains `if __name__ == '__main__'`, throw an error
        """
        with self.assertRaises(prism.exceptions.ParserException) as cm:
            _ = ast_parser.AstParser(IF_NAME_MAIN, MODULE_TEST_CASES)
        expected_msg = f'found `if __name__ == "__main__"` in `{str(IF_NAME_MAIN)}`; all task-specific code should be placed in `run` method'  # noqa: E501
        self.assertEqual(expected_msg, str(cm.exception))

    def test_bad_runs(self):
        """
        Parser throws an error if `run` function is not properly structured or if `run`
        function does not exist
        """

        def _get_args(module):
            parser = ast_parser.AstParser(module, MODULE_TEST_CASES)
            prism_task_class = parser.get_prism_task_node(parser.classes, parser.bases)
            run_func = parser.get_run_func(prism_task_class)
            run_func_args = parser.get_func_args(run_func)
            return run_func_args

        # Extra arg
        run_func_args = _get_args(BAD_RUN_EXTRA_ARG)
        self.assertEqual(
            sorted(['self', 'tasks', 'hooks', 'other_arg']), sorted(run_func_args)
        )

        # Missing arg
        run_func_args = _get_args(BAD_RUN_MISSING_ARG)
        self.assertEqual(sorted(['self', 'tasks']), sorted(run_func_args))

        # No tasks
        run_func_args = _get_args(BAD_RUN_NO_TASKS)
        self.assertEqual(sorted(['self', 'hooks', 'other_arg']), sorted(run_func_args))

        # No run function
        parser = ast_parser.AstParser(NO_RUN_FUNC, MODULE_TEST_CASES)
        prism_task_class = parser.get_prism_task_node(parser.classes, parser.bases)
        run_func = parser.get_run_func(prism_task_class)
        self.assertIsNone(run_func)

        for module in [
            BAD_RUN_EXTRA_ARG, BAD_RUN_MISSING_ARG, BAD_RUN_NO_TASKS, NO_RUN_FUNC
        ]:
            with self.assertRaises(prism.exceptions.ParserException) as cm:
                parser = ast_parser.AstParser(module, MODULE_TEST_CASES)
                parser.parse()
            if module == NO_RUN_FUNC:
                expected_msg = f"no `run` function in PrismTask in `{str(module)}`"
            else:
                expected_msg = f'invalid arguments in `run` function in PrismTask in {str(module)}; should only be `self`,`tasks`,`hooks`'  # noqa: E501
            self.assertEqual(expected_msg, str(cm.exception))



from prism.decorators import task, target
import prism.target
from pathlib import Path


def helper_function_1(tasks):
    return tasks.ref("helper_function1.py")


@task(
    targets=[
        target(type=prism.target.Txt, loc=Path(__file__) / 'test.txt')
    ]
)
def task_with_refs(tasks, hooks):
    _ = tasks.ref('hello.py')
    _ = tasks.ref('world.py')
    return "hi"


def helper_function_2(tasks):
    return tasks.ref("helper_function2.py")



###########
# Imports #
###########

# Prism infrastructure imports
import prism.task
import prism.target
import prism.decorators

# Prism project imports
import prism_project


######################
## Class definition ##
######################

class Module02(prism.task.PrismTask):

    ## Run    
    @prism.decorators.target(type=prism.target.Txt, loc=prism_project.OUTPUT / 'target_txt.txt')
    def run(self, tasks, hooks):
        string = "Hello, world!"
        return string


# EOF



# Standard library imports
import requests
import json

# Prism imports
import prism_project
from prism.decorators import task, target
from prism.target import JSON


# Task
@task(
    targets=[target(type=JSON, loc=prism_project.OUTPUT / 'astros.json')],
)
def extract(tasks, hooks):
    url = "http://api.open-notify.org/astros.json"
    resp = requests.get(url)
    return json.loads(resp.text)



"""
DagCompiler and CompiledDag classes

Table of Contents
- Imports
- Class definition
"""

###########
# Imports #
###########

# Standard library imports
import os
import networkx as nx
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple

# Prism-specific imports
import prism.constants
import prism.exceptions
import prism.prism_logging
import prism.parsers.ast_parser as ast_parser
import prism.infra.module
from prism.infra.manifest import Manifest, ModuleManifest
from prism.infra.project import PrismProject


####################
# Class definition #
####################

class CompiledDag:
    """
    Compiled DAG
    """
    def __init__(self,
        modules_dir: Path,
        nxdag: nx.DiGraph,
        topological_sort: List[Path],
        user_arg_modules: List[Path],
        module_manifests: Dict[Path, ModuleManifest]
    ):
        self.modules_dir = modules_dir
        self.nxdag = nxdag
        self.topological_sort = topological_sort
        self.user_arg_modules = user_arg_modules
        self.module_manifests = module_manifests

        # Store full paths in attribute
        self.topological_sort_full_path = [
            self.modules_dir / module for module in self.topological_sort
        ]

        # Create module objects
        self.compiled_modules = []
        for relative, full in zip(self.topological_sort, self.topological_sort_full_path):  # noqa: E501
            self.compiled_modules.append(
                prism.infra.module.CompiledModule(
                    relative, full, self.module_manifests[relative]
                )
            )


class DagCompiler:
    """
    Class for parsing mod refs, building DAG
    """

    def __init__(self,
        project_dir: Path,
        compiled_dir: Path,
        all_modules: List[Path],
        user_arg_modules: List[Path],
        user_arg_all_downstream: bool,
        project: Optional[PrismProject] = None
    ):
        self.project_dir = project_dir
        self.compiled_dir = compiled_dir
        self.all_modules = all_modules
        self.user_arg_modules = user_arg_modules
        self.project = project
        os.chdir(project_dir)

        # Modules can only be executed if their predecessors are explicitly run or have
        # targets. For example, if our DAG is A --> B --> C and we call `prism run
        # --module C`, then Prism will parse the execution order, instantiate but NOT
        # execute tasks A and B, and then run task C. In other words, A and B will
        # always be instantiated; the --all-upstream argument controls whether A and B
        # are executed.

        # This is why we don't need to keep track of the --all-upstream argument; it
        # doesn't affect what our topological sort looks like. However, the
        # --all-downstream argument does, since successors do not need to be
        # instantiated at all.
        self.user_arg_all_downstream = user_arg_all_downstream

        # Path of modules
        self.modules_dir = self.project_dir / 'modules'

        # Module manifests
        self.module_manifests: Dict[Path, ModuleManifest] = {}

    def parse_task_refs(self,
        modules: List[Path],
        parent_path: Path
    ) -> Dict[Path, Any]:
        """
        Parse node dictionary listed at the beginning of each python script. If
        node_dict does not exist in any script, throw an error.

        args:
            modules: modules to compile
            parent_path: parent path of modules
        returns:
            module references as a dictionary
        """

        # This is only ever called on the output of `get_all_modules`, which sorts the
        # modules alphabetically. Therefore, all mod refs will be sorted.
        task_refs_dict: Dict[Path, Any] = {}
        for m in modules:
            parser = ast_parser.AstParser(m, parent_path)
            task_refs = parser.parse()
            if task_refs is None or task_refs == '' or task_refs == []:
                task_refs_dict[m] = None
            else:
                task_refs_dict[m] = task_refs

            # Keep track of module manifest
            self.module_manifests[m] = parser.module_manifest

        return task_refs_dict

    def add_graph_elem(self,
        elem: Any,
        master: List[Any]
    ) -> List[Any]:
        """
        Add graph element `elem` to master list of elements

        args:
            elem: graph element to add (either a node or an edge)
            master: master list of elements (either nodes or edges)
        returns
            master list of elements with new elem
        """
        if elem not in master:
            master += [elem]  # type: ignore
        else:
            return master
        return master

    def add_graph_node(self,
        elem: Path,
        master: List[Path]
    ) -> List[Path]:
        """
        To resolve mypy errors...
        """
        return self.add_graph_elem(elem, master)

    def add_graph_edge(self,
        elem: Tuple[Path, Path],
        master: List[Tuple[Path, Path]]
    ) -> List[Tuple[Path, Path]]:
        """
        To resolve mypy errors...
        """
        return self.add_graph_elem(elem, master)

    def create_nodes_edges(self,
        module_references: Dict[Path, Any]
    ) -> Tuple[List[Path], List[Tuple[Path, Path]]]:
        """
        Create nodes / edges from module connections

        args:
            module_references: connections defined via {{ mod(...) }} in modules
        outputs:
            nodes: list of nodes (modules)
            edges: list of edges (tuple of nodes, i.e. modules)
        """

        # Create edges and nodes
        edges: List[Tuple[Path, Path]] = []
        nodes: List[Path] = []

        # Iterate through module references. Keys represent distinct modules in the DAG,
        # and values represent the modules that feed into the key.
        for mod, ref in module_references.items():
            nodes = self.add_graph_node(mod, nodes)
            if ref is None:
                pass
            elif isinstance(ref, Path):
                nodes = self.add_graph_node(ref, nodes)
                edges = self.add_graph_edge((ref, mod), edges)
            else:
                for v in ref:
                    nodes = self.add_graph_node(v, nodes)
                    edges = self.add_graph_edge((v, mod), edges)
        return nodes, edges

    def create_dag(self,
        nodes: List[Path],
        edges: List[Tuple[Path, Path]]
    ) -> nx.DiGraph:
        """
        Create DAG from edges

        args:
            user_arg_modules: modules passed in user arguments
            nodes: list of nodes (modules)
            edges: list of edges (tuple of nodes, i.e. modules)
        outputs:
            topological sort of edges
        """
        # Instantiate graph
        graph = nx.DiGraph()

        # If there are no nodes in the DAG, then throw an error. This should never
        # happen.
        if len(nodes) == 0:
            raise prism.exceptions.RuntimeException(message='DAG has 0 nodes')

        # Add the edges/nodes
        graph.add_edges_from(edges)
        graph.add_nodes_from(nodes)

        # Check if graph is a DAG
        if not nx.is_directed_acyclic_graph(graph):
            all_scc = nx.algorithms.components.strongly_connected_components(graph)
            cycles = []
            for component in all_scc:
                component = set([str(elem) for elem in list(component)])
                if len(component) > 1:
                    cycles.append(component)
            raise prism.exceptions.DAGException(
                message=f"invalid DAG, cycle found in {str(cycles)}"
            )

        return graph

    def get_node_dependencies(self,
        graph: nx.DiGraph,
        start_nodes: List[Path]
    ) -> List[Path]:
        """
        Parse the DiGraph and get all nodes upstream of the `start_nodes`

        args:
            graph: DAG
            start_nodes: list of nodes for which to retrieve upstream dependencies
        returns:
            list of dependencies
        """
        # Add all start nodes to the list of dependencies with a depth of 0
        deps = [x for x in start_nodes]

        # Iterate through the start nodes and extract the predecessors
        for node in deps:
            for prenode in graph.predecessors(node):
                deps.append(prenode)

        # Get unique dependencies
        unique_deps = list(set(deps))
        return unique_deps

    def get_node_successors(self,
        graph: nx.DiGraph,
        start_nodes: List[Path]
    ) -> List[Path]:
        """
        Parse the DiGraph and get all nodes downstream of the `start_nodes`

        args:
            graph: DAG
            start_nodes: list of nodes for which to retrieve upstream dependencies
        returns:
            list of successors
        """
        # Add all start nodes to the list of successors with a depth of 0
        successors = [x for x in start_nodes]

        # Iterate through the start nodes and extract the predecessors
        for node in successors:
            for postnode in graph.successors(node):
                successors.append(postnode)

        # Get unique dependencies
        unique_successors = list(set(successors))
        return unique_successors

    def create_topsort(self,
        all_modules: List[Path],
        user_arg_modules: List[Path],
        parent_path: Path
    ) -> Any:
        """
        Parse mod refs, create the DAG, and create a topological sort of the DAG

        args:
            all_modules: list of all modules
            user_arg_modules: modules passed in user arguments
            parent_path: path containing the modules
            compiler_dict: globals dictionary for compiler
        returns:
            topological sorted DAG as a list
        """

        # Create a DAG using all the modules. We use `all_modules` instead of
        # `user_arg_modules`, because using `user_arg_modules` will only compile/run the
        # modules referenced in the modules themselves. For example, if we have a dag A
        # --> B --> C and wish to to only compile/run script C, then our code will only
        # run script B. This will throw an error, because script B relies on script A,
        # and we will need to instantiate the script A task for the script B task to
        # execute fully.
        task_refs = self.parse_task_refs(all_modules, parent_path)
        nodes, edges = self.create_nodes_edges(task_refs)
        dag = self.create_dag(nodes, edges)

        # If `user_arg_modules` is equivalent to `all_modules`, then create a
        # topological sorting of the full DAG. From the NetworkX documentation: A
        # topological sort is a nonunique permutation of the nodes of a directed graph
        # such that an edge from u to v implies that u appears before v in the
        # topological sort order. This ordering is valid only if the graph has no
        # directed cycles.
        if len(user_arg_modules) == len(all_modules):
            all_topological_sorts = nx.algorithms.dag.all_topological_sorts(dag)
            all_topological_sorts_list = next(all_topological_sorts)

        # Otherwise, the user has selected to run a subset of the modules. Identify all
        # modules upstream (and potentially downstream) of `user_arg_modules`.
        else:

            # Keep only the dependencies, and create a topological sort.
            all_nodes = self.get_node_dependencies(dag, user_arg_modules)

            # Add successors if the user wants them
            if self.user_arg_all_downstream:
                all_nodes.extend(self.get_node_successors(dag, user_arg_modules))

            subgraph = dag.subgraph(list(set(all_nodes)))
            all_topological_sorts = nx.algorithms.dag.all_topological_sorts(subgraph)  # noqa: E501
            all_topological_sorts_list = next(all_topological_sorts)

        # Add each module to manifest
        for elem in all_topological_sorts_list:

            # Raise error if node not in project
            if elem not in all_modules:
                raise prism.exceptions.CompileException(
                    message=f'module `{str(elem)}` not found in project'
                )

        return dag, all_topological_sorts_list

    def compile(self) -> CompiledDag:
        """
        Compile the DAG
        """
        nxdag, all_topological_sorts_list = self.create_topsort(
            self.all_modules, self.user_arg_modules, self.modules_dir
        )

        # Dump manifest
        manifest = Manifest(list(self.module_manifests.values()))

        # Add the prism project to the Manifest
        if self.project is not None:
            prism_project_py_str = self.project.prism_project_py_str
        else:
            prism_project = PrismProject(
                project_dir=self.project_dir,
                user_context={},
                which="compile",
                filename="prism_project.py"
            )
            prism_project_py_str = prism_project.prism_project_py_str
        manifest.add_prism_project(prism_project_py_str)
        manifest.json_dump(self.compiled_dir)

        # Return dag
        dag = CompiledDag(
            self.modules_dir,
            nxdag,
            all_topological_sorts_list,
            self.user_arg_modules,
            self.module_manifests
        )
        return dag



"""
Prism Manifest class

Table of Contents
- Imports
- Class definition
"""

###########
# Imports #
###########

# Standard library imports
import json
from pathlib import Path
from typing import Any, Dict, List, Union


####################
# Class definition #
####################

class ModuleManifest:
    """
    Class used to store metadata on a parsed module
    """

    def __init__(self):
        self.manifest_dict: Dict[str, Any] = {"targets": [], "modules": [], "refs": []}

    def add_module(self, module_name: Path):
        self.manifest_dict["modules"].append(str(module_name))

    def add_ref(self, target: Path, source: Path):
        obj = {
            "target": str(target),
            "source": str(source)
        }
        self.manifest_dict["refs"].append(obj)

    def add_target(self, module_name: Path, loc: Union[str, List[str]]):
        obj = {
            "module_name": str(module_name),
            "target_locs": loc
        }
        self.manifest_dict["targets"].append(obj)


class Manifest:
    """
    Class used to store metadata on compiled prism project
    """

    def __init__(self, module_manifests: List[ModuleManifest] = []):
        self.manifest_dict: Dict[str, Any] = {
            "targets": [], "prism_project": "", "modules": [], "refs": []
        }
        self.module_manifests = module_manifests

        # Iterate through module manifests and add to manifest
        for mm in self.module_manifests:
            self.manifest_dict["targets"].extend(mm.manifest_dict["targets"])
            self.manifest_dict["modules"].extend(mm.manifest_dict["modules"])
            self.manifest_dict["refs"].extend(mm.manifest_dict["refs"])

    def add_prism_project(self, prism_project_data: str):
        self.manifest_dict["prism_project"] = prism_project_data

    def add_module(self, module_name: Path, module_data: str):
        obj = {
            "module_name": str(module_name),
            "module_data": module_data
        }
        self.manifest_dict["modules"].append(obj)

    def add_ref(self, target: Path, source: Path):
        obj = {
            "target": str(target),
            "source": str(source)
        }
        self.manifest_dict["refs"].append(obj)

    def add_target(self, module_name: Path, loc: Union[str, List[str]]):
        obj = {
            "module_name": str(module_name),
            "target_locs": loc
        }
        self.manifest_dict["targets"].append(obj)

    def json_dump(self, path: Path):
        with open(path / 'manifest.json', 'w') as f:
            json.dump(self.manifest_dict, f, sort_keys=False)
        f.close()

    def json_load(self, path: Path):
        with open(path / 'manifest.json', 'r') as f:
            manifest = json.loads(f.read())
        f.close()
        return manifest




from pathlib import Path
TASK_REF_15NODES_LIST = [
    Path('module01.py'),
    Path('module02.py'),
    Path('module03.py'),
    Path('module04.py'),
    Path('module05.py'),
    Path('module06.py'),
    Path('module07.py'),
    Path('module08.py'),
    Path('module09.py'),
    Path('module10.py'),
    Path('module11.py'),
    Path('module12.py'),
    Path('module13.py'),
    Path('module14.py'),
    Path('module15.py'),
]


# EOF


###########
# Imports #
###########

# Prism infrastructure imports
import prism.task
import prism.target
import prism.decorators

# Prism project imports
import prism_project


######################
## Class definition ##
######################

class Module02(prism.task.PrismTask):

    ## Run
    def run(self, tasks, hooks):
        """
        Execute task.

        args:
            tasks: used to reference output of other tasks --> tasks.ref('...')
            hooks: built-in Prism hooks. These include:
                - hooks.dbt_ref --> for getting dbt models as a pandas DataFrame
                - hooks.sql     --> for executing sql query using an adapter in profile YML
                - hooks.spark   --> for accessing SparkSession (if pyspark specified in profile YML)
        returns:
            task output
        """
        return tasks.ref('module03.py') + '\n' + 'Hello fro module 3!'


# EOF


"""
Unit testing for parsing modules with tasks created via decorated functions.

Table of Contents:
- Imports
- Test case directory and paths
- Test case class definition
"""

###########
# Imports #
###########

# Standard library imports
import os
from pathlib import Path
import unittest

# Prism imports
import prism.exceptions
import prism.parsers.ast_parser as ast_parser


#################################
# Test case directory and paths #
#################################

# Directory containing all prism_project.py test cases
TEST_CASE_WKDIR = os.path.dirname(__file__)
MODULE_TEST_CASES = Path(TEST_CASE_WKDIR) / 'test_modules'


# Tasks with decorators
DEC_BAD_DEC_NO_PARENTHESES = Path('dec_bad_dec_no_parentheses.py')
DEC_BAD_RUN_EXTRA_ARG = Path('dec_bad_run_extra_arg.py')
DEC_BAD_RUN_MISSING_ARG = Path('dec_bad_run_missing_arg.py')
DEC_DIFF_DECORATOR_STRUCTURE = Path('dec_diff_decorator_structure.py')
DEC_MULTIPLE_PRISM_TASKS = Path('dec_multiple_prism_tasks.py')
DEC_NO_PRISM_TASK = Path('dec_no_prism_task.py')
DEC_ONE_PRISM_TASK = Path('dec_one_prism_task.py')
DEC_OTHER_FUNCTIONS = Path('dec_other_functions.py')
DEC_TASK_WITH_TARGET = Path('dec_task_with_target.py')
DEC_TASKS_REFS = Path('dec_tasks_refs.py')


##############################
# Test case class definition #
##############################

class TestModuleDecParsing(unittest.TestCase):

    def test_one_prism_task(self):
        """
        Test behavior of parser with just one Prism task
        """
        # Prism task
        parser = ast_parser.AstParser(DEC_ONE_PRISM_TASK, MODULE_TEST_CASES)

        # Prism task name
        prism_task_fn = parser.get_prism_task_node(parser.classes, parser.bases)
        prism_task_name = prism_task_fn.name
        prism_task_fn_args = parser.get_func_args(prism_task_fn)

        # Number of prism tasks
        num_prism_tasks = parser.get_num_prism_task_functions()

        self.assertEqual("task_function", prism_task_name)
        self.assertEqual(1, num_prism_tasks)
        self.assertEqual(['tasks', 'hooks'], prism_task_fn_args)

        # Calling `parse` shouldn't throw an error
        self.assertEqual([], parser.parse())

    def test_no_prism_task(self):
        """
        Test behavior of parser when module has no Prism task
        """
        parser = ast_parser.AstParser(DEC_NO_PRISM_TASK, MODULE_TEST_CASES)

        # Prism task name
        prism_task_fn = parser.get_prism_task_node(parser.classes, parser.bases)

        # Number of prism tasks
        num_prism_tasks = parser.get_num_prism_task_functions()

        self.assertIsNone(prism_task_fn)
        self.assertEqual(0, num_prism_tasks)

        # Calling `parse` should throw an error
        with self.assertRaises(prism.exceptions.ParserException) as cm:
            parser.parse()
        expected_msg = f"no PrismTask in `{str(DEC_NO_PRISM_TASK)}`"
        self.assertEqual(expected_msg, str(cm.exception))

    def test_multiple_prism_task(self):
        """
        Test behavior of parser when module has multiple Prism tasks
        """
        parser = ast_parser.AstParser(DEC_MULTIPLE_PRISM_TASKS, MODULE_TEST_CASES)

        # Number of prism tasks
        num_prism_tasks = parser.get_num_prism_task_functions()
        self.assertEqual(2, num_prism_tasks)

        # Prism task name -- it should return the first one
        with self.assertRaises(prism.exceptions.RuntimeException) as cm:
            _ = parser.get_prism_task_node(
                parser.classes, parser.bases
            )
        expected_msg = f"too many functions decorated with `@task` in {DEC_MULTIPLE_PRISM_TASKS}"  # noqa: E501
        self.assertEqual(expected_msg, str(cm.exception))

    def test_diff_decorator_structure(self):
        """
        Different @task decorator structure (i.e., import @prism.decorators.task())
        should not affect the behavior of the parser
        """
        # Prism task
        parser = ast_parser.AstParser(DEC_DIFF_DECORATOR_STRUCTURE, MODULE_TEST_CASES)

        # Prism task function
        prism_task_fn = parser.get_prism_task_node(parser.classes, parser.bases)
        prism_fn_name = prism_task_fn.name
        run_func_args = parser.get_func_args(prism_task_fn)

        # Number of prism tasks
        num_prism_tasks = parser.get_num_prism_task_functions()

        self.assertEqual("task_fn_different_decorator_structure", prism_fn_name)
        self.assertEqual(1, num_prism_tasks)
        self.assertEqual(['tasks', 'hooks'], run_func_args)

        # Calling `parse` shouldn't throw an error
        self.assertEqual([], parser.parse())

    def test_other_functions(self):
        """
        Presence of other, non-PrismTask classes should not affect the behavior of the
        parser
        """
        # Prism task
        parser = ast_parser.AstParser(DEC_OTHER_FUNCTIONS, MODULE_TEST_CASES)

        # Prism task name
        prism_task_fn = parser.get_prism_task_node(parser.classes, parser.bases)
        prism_task_name = prism_task_fn.name
        run_func_args = parser.get_func_args(prism_task_fn)

        # Number of prism tasks
        num_prism_tasks = parser.get_num_prism_task_functions()

        self.assertEqual("task_function", prism_task_name)
        self.assertEqual(1, num_prism_tasks)
        self.assertEqual(['tasks', 'hooks'], run_func_args)

        # Calling `parse` shouldn't throw an error
        self.assertEqual([], parser.parse())

    def test_task_with_target(self):
        """
        Presence of target does not affect parser behavior
        """
        # Prism task
        parser = ast_parser.AstParser(DEC_TASK_WITH_TARGET, MODULE_TEST_CASES)

        # Prism task name
        prism_task_fn = parser.get_prism_task_node(parser.classes, parser.bases)
        prism_task_name = prism_task_fn.name
        run_func_args = parser.get_func_args(prism_task_fn)

        # Number of prism tasks
        num_prism_tasks = parser.get_num_prism_task_functions()

        self.assertEqual("task_with_target", prism_task_name)
        self.assertEqual(1, num_prism_tasks)
        self.assertEqual(['tasks', 'hooks'], run_func_args)

        # Calling `parse` shouldn't throw an error
        self.assertEqual([Path('hello.py'), Path('world.py')], parser.parse())

        # Get target
        targets = parser.get_targets(prism_task_fn, prism_task_fn)
        expected_targets = "Path(__file__) / 'test.txt'"
        self.assertEqual(targets, expected_targets)

    def test_tasks_refs(self):
        """
        Test behavior of parse when there are mod references
        """
        # Prism task
        parser = ast_parser.AstParser(DEC_TASKS_REFS, MODULE_TEST_CASES)

        # Prism task name
        prism_task_fn = parser.get_prism_task_node(parser.classes, parser.bases)
        prism_task_name = prism_task_fn.name
        run_func_args = parser.get_func_args(prism_task_fn)

        # Number of prism tasks
        num_prism_tasks = parser.get_num_prism_task_functions()

        self.assertEqual("task_with_refs", prism_task_name)
        self.assertEqual(1, num_prism_tasks)
        self.assertEqual(['tasks', 'hooks'], run_func_args)

        # Calling `parse` shouldn't throw an error
        expected_tasks = [
            Path('helper_function1.py'),
            Path('helper_function2.py'),
            Path('hello.py'),
            Path('world.py')
        ]
        self.assertEqual(sorted(expected_tasks), sorted(parser.parse()))

    def test_bad_function_dec_definition(self):
        """
        Parser throws an error if:
            1. The function decorated with `@task` is not properly structured, or
            2. The @task decorator is not properly specified
        """

        def _get_args(module):
            parser = ast_parser.AstParser(module, MODULE_TEST_CASES)
            prism_task_fn = parser.get_prism_task_node(parser.classes, parser.bases)
            run_func_args = parser.get_func_args(prism_task_fn)
            return run_func_args

        # Extra arg
        run_func_args = _get_args(DEC_BAD_RUN_EXTRA_ARG)
        self.assertEqual(
            sorted(['tasks', 'hooks', 'other_arg']), sorted(run_func_args)
        )

        # Missing arg
        run_func_args = _get_args(DEC_BAD_RUN_MISSING_ARG)
        self.assertEqual(sorted(['tasks']), sorted(run_func_args))

        # @task de
        run_func_args = _get_args(DEC_BAD_DEC_NO_PARENTHESES)
        self.assertEqual(sorted(['tasks', 'hooks']), sorted(run_func_args))

        # Missing arguments raise a ParserException
        for module in [
            DEC_BAD_RUN_EXTRA_ARG, DEC_BAD_RUN_MISSING_ARG,
        ]:
            with self.assertRaises(prism.exceptions.ParserException) as cm:
                parser = ast_parser.AstParser(module, MODULE_TEST_CASES)
                parser.parse()
            expected_msg = f'invalid arguments in `run` function in PrismTask in {str(module)}; should only be `tasks`,`hooks`'  # noqa: E501
            self.assertEqual(expected_msg, str(cm.exception))

        # An improper @task decorator raises a RuntimeException
        with self.assertRaises(prism.exceptions.RuntimeException) as cm:
            parser = ast_parser.AstParser(DEC_BAD_DEC_NO_PARENTHESES, MODULE_TEST_CASES)
            parser.parse()
        expected_msg = "`task` decorator not properly specified...try adding parentheses to it, e.g., `@task()`"  # noqa: E501
        self.assertEqual(expected_msg, str(cm.exception))



from prism.task import PrismTask

class TasksRefs(PrismTask):
    
    def func_0(self, tasks, hooks):
        return tasks.ref('func_0.py')

    def run(self, tasks, hooks):
        x = tasks.ref('hello.py')
        y = tasks.ref('world.py')
        return 'hi'

    def func_1(self, tasks, hooks):
        return tasks.ref('func_1.py')


# EOF


"""
Prism project
"""

# Imports
import logging
from pathlib import Path
from prism.admin import generate_run_id, generate_run_slug


# Project metadata
NAME = ""
AUTHOR = ""
VERSION = ""
DESCRIPTION = """
"""

# Admin
RUN_ID = generate_run_id()  # don't delete this!
SLUG = generate_run_slug()  # don't delete this!


# sys.path config. This gives your tasks access to local modules / packages that exist
# outside of your project structure.
SYS_PATH_CONF = [
    Path(__file__).parent,
    Path(__file__).parent.parent,
]


# Thread count: number of workers to use to execute tasks concurrently. If set to 1,
# then 1 task is run at a time.
THREADS = 2


# Profile directory and name
PROFILE_YML_PATH = Path(__file__).parent / 'profile.yml'
PROFILE = None  # name of profile within `profiles.yml`


# Logger
PRISM_LOGGER = logging.getLogger("PRISM_LOGGER")


# Other variables / parameters. Make sure to capitalize all of these!
VAR_1 = {'a': 'b'}
VAR_2 = 200
VAR_3 = '2015-01-01'

# Paths
WKDIR = Path(__file__).parent
DATA = WKDIR / 'data'
OUTPUT = WKDIR / 'output'



"""
Mixin classes for compile task

Table of Contents
- Imports
- Class definition
"""


###########
# Imports #
###########

# Standard library imports
import re
import argparse
from pathlib import Path
from typing import List, Optional

# Prism-specific imports
import prism.cli.base
import prism.exceptions
import prism.constants
from prism.infra import compiler
from prism.infra.project import PrismProject


####################
# Class definition #
####################

class CompileMixin():
    """
    Mixin for compile task
    """

    def is_valid_script_name(self,
        script: str
    ) -> bool:
        """
        Check if script name is valid. Script must start with [A-Za-z] character to be
        valid.

        args:
            script: script name
        returns:
            prism.exceptions.CompileException
        """
        if len(re.findall(r'\.py$', script)) == 0:
            pass
        else:
            if len(re.findall(r'(?i)^[a-z]', script)) > 0:
                return True
        raise prism.exceptions.CompileException(
            message=f'`{script}` is not valid name; script must start with [a-zA-Z] to be compiled correctly'  # noqa: E501
        )

    def get_modules_dir(self,
        project_dir: Path
    ) -> Path:
        """
        Get directory containing modules

        args:
            project_dir: project directory
        returns:
            modules directory
        """
        modules_dir = project_dir / 'modules'
        if not modules_dir.is_dir():
            raise prism.exceptions.CompileException(
                message=f'`modules` directory not found in `{str(project_dir)}`'
            )
        return modules_dir

    def get_modules(self,
        modules_dir: Path,
        prefix: Path = Path('')
    ) -> List[Path]:
        """
        Get all Python modules from `modules_dir`

        args:
            modules_dir: modules directory
            prefix: prefix to use for directories in modules directory
        returns:
            list of modules in directory (as pathlib Path objects)
        """
        modules = []
        for path in modules_dir.iterdir():

            # If object is a directory...
            if path.is_dir():

                # If parent directory is the modules folder, set the prefix to the name
                # of the directory. For example, if the modules folder has a directory
                # called "extraction", then the modules within "extraction" should be
                # stored as "extraction/..."
                if str(prefix) == '.':
                    modules.extend(self.get_modules(path, prefix=Path(path.name)))

                # If parent directory is not the modules folder, set the prefix to the
                # parent prefix + the name of the directory. Using the above example, if
                # the "extraction" folder has a directory called "step1", then the
                # modules within "step1" should be stored as "extraction/step1/..."
                else:
                    modules.extend(
                        self.get_modules(path, prefix=Path(prefix) / path.name)
                    )

            # If object is a file
            elif path.is_file():

                # If file is a python file...
                if len(re.findall(r'\.py$', path.name)) > 0:

                    # If parent directory is the modules folder, then just add the
                    # python file name
                    if str(prefix) == '.':
                        modules += [Path(path.name)]

                    # If parent directory is not the modules folder, then add prefix to
                    # python file name
                    else:
                        modules += [Path(prefix) / Path(path.name)]
        return modules

    def user_arg_modules(self,
        args: argparse.Namespace,
        modules_dir: Path
    ) -> List[Path]:
        """
        Process user arg modules

        args:
            args: user arguments
            modules_dir: modules directory
        returns:
            list of processed user arg modules
        """
        try:
            raw_modules = args.modules
            if raw_modules is None:
                processed_modules = self.get_modules(modules_dir)
            else:
                processed_modules = []
                for m in raw_modules:

                    # Check if * is used
                    m_split = m.split('/')
                    if m_split[-1] == '*':
                        parent = '/'.join([m_comp for m_comp in m_split[:-1]])
                        processed_modules.extend(
                            self.get_modules(Path(modules_dir / parent), Path(parent))
                        )

                    # Check if path is a directory
                    elif Path(modules_dir / m).is_dir():
                        raise prism.exceptions.CompileException(
                            message=f'invalid --module argument `{m}`'
                        )

                    # Check if path is a file
                    elif Path(modules_dir / m).is_file():
                        if len(re.findall(r'\.py$', Path(modules_dir / m).name)) == 0:
                            raise prism.exceptions.CompileException(
                                f'invalid --module argument `{m}`'
                            )
                        else:
                            processed_modules.extend([Path(m)])

                    else:
                        raise prism.exceptions.CompileException(
                            f'could not find module `{str(m)}` in `{str(modules_dir.name)}/`'  # noqa: E501
                        )

        # If --module argument not specified, then get all modules
        except AttributeError:
            processed_modules = self.get_modules(modules_dir)

        return processed_modules

    def get_compiled_dir(self,
        project_dir: Path
    ) -> Path:
        """
        Get directory containing compiled modules

        args:
            project_dir: project directory
        returns:
            compiled_dir: directory with compiled modules
        """
        compiled_dir = project_dir / '.compiled'
        return compiled_dir

    def create_compiled_dir(self,
        project_dir: Path
    ):
        """
        Create folder for compiled modules in project directory

        args:
            project_dir: project directory
            modules: list of modules to compile
        returns:
            None
        """

        # Create compiled directory
        compiled_modules_path = self.get_compiled_dir(project_dir)
        if not compiled_modules_path.is_dir():
            compiled_modules_path.mkdir(parents=True, exist_ok=True)
        return compiled_modules_path

    def compile_dag(self,
        project_dir: Path,
        compiled_dir: Path,
        all_modules: List[Path],
        user_arg_modules: List[Path],
        user_arg_all_downstream: bool = True,
        project: Optional[PrismProject] = None
    ) -> compiler.CompiledDag:
        """
        Wrapper for the `compile` method in the DagCompiler class

        args:
            project_dir: project directory
            all_modules: all modules in project
            user_arg_modules: modules passed in user argument
            user_arg_all_downstream: boolean indicating whether the user wants to run
                all modules downstream of inputted args
            compiler_globals: globals() dictionary
        returns:
            CompiledDag object
        """
        dag_compiler = compiler.DagCompiler(
            project_dir,
            compiled_dir,
            all_modules,
            user_arg_modules,
            user_arg_all_downstream,
            project
        )
        compiled_dag = dag_compiler.compile()

        # Check that all modules in topological sort are in project
        for m in compiled_dag.topological_sort:
            if m not in all_modules:
                raise prism.exceptions.CompileException(
                    message=f'module `{m}` not found in project'
                )

        # Otherwise, return
        return compiled_dag
